
#include <allegro5/allegro.h>
#include <allegro5/allegro_image.h>
#include <allegro5/allegro_primitives.h>
#include <allegro5/allegro_font.h>
#include <allegro5/allegro_ttf.h>

#include <cstdio>
#include <cmath>
#include <exception>

using std::exception;
///typedef std::exception GameException;

class GameException : public exception {
   const char* str;
public :
   GameException(const char* s) : str(s) {}
   
   const char* what() {return str;}
};

class POINT {
public :
   float x;
   float y;
   
   POINT() : x(0.0f) , y(0.0f) {}
   POINT(float xpos , float ypos) : x(xpos) , y(ypos) {}
   
   POINT& SetPoint(float xpos , float ypos) {
      x = xpos;
      y = ypos;
      return *this;
   }
   
   POINT operator-() {
      POINT r;
      r.x = -x;
      r.y = -y;
      return r;
   }
   POINT operator+(POINT& p) {
      POINT r = *this;
      r.x += p.x;
      r.y += p.y;
      return r;
   }
   POINT& operator+=(POINT& p) {
      *this = *this + p;
      return *this;
   }
   POINT operator-(POINT& p) {
      POINT r = *this;
      r.x -= p.x;
      r.y -= p.y;
      return r;
   }
   POINT& operator-=(POINT& p) {
      *this = *this - p;
      return *this;
   }
   
   friend bool OnSameTile(POINT p1 , POINT p2);
};

bool OnSameTile(POINT p1 , POINT p2) {
   return ((floor(p1.x) == floor(p2.x)) && (floor(p1.y) == floor(p2.y)));
}

class IsometricWorld {
   
private :
   ALLEGRO_DISPLAY* display;
   
   ALLEGRO_EVENT_QUEUE* queue;
   ALLEGRO_TIMER* timer;
   
   ALLEGRO_FONT* font;
   
   ALLEGRO_BITMAP* tile1;
   ALLEGRO_BITMAP* tile2;
   
   float SCREENWIDTH;
   float SCREENHEIGHT;
   
   float NUMTILESWIDE;
   float NUMTILESTALL;
   
   float ISOTILEWIDTH;
   float ISOTILEHEIGHT;
   
   POINT l,r,t,b,c;
   
   POINT camPos;
   POINT isoWorldPos;
   
   POINT msPos;
   
   
public :
   
   IsometricWorld() :
         display(0),
         queue(0),
         timer(0),
         font(0),
         tile1(0),
         tile2(0),
         SCREENWIDTH(800),
         SCREENHEIGHT(600),
         NUMTILESWIDE(4),
         NUMTILESTALL(4),
         ISOTILEWIDTH(SCREENWIDTH/NUMTILESWIDE),
         ISOTILEHEIGHT(SCREENHEIGHT/NUMTILESTALL),
         l(),r(),t(),b(),c(),
         camPos(0.0f , 0.0f),
         isoWorldPos(SCREENWIDTH/2.0f , SCREENHEIGHT)
   {
      if (!al_init())                   {throw GameException("Failed to initialize allegro 5.");}
      if (!al_init_image_addon())       {throw GameException("Failed to init image addon.");}
      if (!al_init_primitives_addon())  {throw GameException("Failed to init primitives addon.");}
      if (!al_init_font_addon())        {throw GameException("Failed to init font addon.");}
      if (!al_init_ttf_addon())         {throw GameException("Failed to init ttf addon.");}
      if (!al_install_keyboard())       {throw GameException("Failed to install keyboard.");}
      if (!al_install_mouse())          {throw GameException("Failed to install mouse.");}
      
      al_set_new_display_flags(ALLEGRO_OPENGL | ALLEGRO_WINDOWED);
      display = al_create_display(SCREENWIDTH , SCREENHEIGHT);
      
      if (!display) {throw GameException("Failed to create display.");}
      
      al_clear_to_color(al_map_rgb(0,0,0));
      al_flip_display();
      
      timer = al_create_timer(1.0/60.0);
      if (!timer) {throw GameException("Failed to create timer.");}
      
      queue = al_create_event_queue();
      if (!queue) {throw GameException("Failed to create event queue.");}
      
      al_register_event_source(queue , al_get_display_event_source(display));
      al_register_event_source(queue , al_get_keyboard_event_source());
      al_register_event_source(queue , al_get_mouse_event_source());
      
      font = al_load_ttf_font("verdana.ttf" , 24 , 0);
      if (!font) {throw GameException("Failed to load font.");}
      
      tile1 = al_create_bitmap(ISOTILEWIDTH , ISOTILEHEIGHT);
      tile2 = al_create_bitmap(ISOTILEWIDTH , ISOTILEHEIGHT);
      
      l.SetPoint(0.0f , ISOTILEHEIGHT/2.0f);
      r.SetPoint(ISOTILEWIDTH , ISOTILEHEIGHT/2.0f);
      t.SetPoint(ISOTILEWIDTH/2.0f , 0.0f);
      b.SetPoint(ISOTILEWIDTH/2.0f , ISOTILEHEIGHT);
      c.SetPoint(ISOTILEWIDTH/2.0f , ISOTILEHEIGHT/2.0f);
      
      
      ALLEGRO_COLOR c1 = al_map_rgb(255,192,0);
      ALLEGRO_COLOR c2 = al_map_rgb(0,64,255);
      float thickness = 10.0;
      
      al_set_target_bitmap(tile1);
      
      al_clear_to_color(al_map_rgba(0,0,0,0));
      al_draw_line(l.x , l.y , t.x , t.y , c1 , thickness);
      al_draw_line(l.x , l.y , b.x , b.y , c1 , thickness);
      al_draw_line(r.x , r.y , t.x , t.y , c1 , thickness);
      al_draw_line(r.x , r.y , b.x , b.y , c1 , thickness);
      
      al_set_target_bitmap(tile2);
      
      al_clear_to_color(al_map_rgba(0,0,0,0));
      al_draw_line(l.x , l.y , t.x , t.y , c1 , thickness);
      al_draw_line(l.x , l.y , b.x , b.y , c1 , thickness);
      al_draw_line(r.x , r.y , t.x , t.y , c1 , thickness);
      al_draw_line(r.x , r.y , b.x , b.y , c1 , thickness);
      al_draw_filled_triangle(l.x , l.y , t.x , t.y , c.x , c.y , c2);
      al_draw_filled_triangle(r.x , r.y , t.x , t.y , c.x , c.y , c2);
      al_draw_filled_triangle(l.x , l.y , b.x , b.y , c.x , c.y , c2);
      al_draw_filled_triangle(r.x , r.y , b.x , b.y , c.x , c.y , c2);
      
      al_set_target_bitmap(al_get_backbuffer(display));
   }
   
   void DrawBitmapCenteredOn(ALLEGRO_BITMAP* bmp , POINT p) {
      p.x -= al_get_bitmap_width(bmp)/2.0f;
      p.y -= al_get_bitmap_height(bmp)/2.0f;
      
      al_draw_bitmap(bmp , p.x , p.y , 0);
   }
   
   void Run() {
      al_start_timer(timer);
      bool redraw = true;
      
      while(1) {
         if (redraw) {
            al_clear_to_color(al_map_rgb(0,0,0));
            POINT isoMsPos = ScreenCoordsToIsoCoords(msPos);
            for (unsigned int tiley = 0 ; tiley < NUMTILESTALL ; ++tiley) {
               for (unsigned int tilex = 0 ; tilex < NUMTILESWIDE ; ++tilex) {
                  POINT isoPt((float)tilex + 0.5f , (float)tiley + 0.5f);
                  POINT worldPt = IsoCoordsToScreenCoords(isoPt);
                  
                  if (OnSameTile(isoPt , isoMsPos)) {
                     DrawBitmapCenteredOn(tile2 , worldPt);
                  }
                  else {
                     DrawBitmapCenteredOn(tile1 , worldPt);
                  }
               }
            }
            
            al_draw_textf(font , al_map_rgb(255,255,255) , 10 , 10 , 0 , "   msPos = %3.1f , %3.1f" , msPos.x , msPos.y);
            al_draw_textf(font , al_map_rgb(255,255,255) , 10 , 50 , 0 , "isoMsPos = %3.1f , %3.1f" , isoMsPos.x , isoMsPos.y);
            
            al_flip_display();
            redraw = false;
         }
         
         do {
            ALLEGRO_EVENT ev;
            al_wait_for_event(queue , &ev);
            if (ev.type == ALLEGRO_EVENT_DISPLAY_CLOSE) {return;}
            
            if (ev.type == ALLEGRO_EVENT_MOUSE_AXES) {
               msPos.SetPoint(ev.mouse.x , ev.mouse.y);
               redraw = true;
            }
            else if (ev.type == ALLEGRO_EVENT_TIMER) {
               redraw = true;
            }
            
            
            if (ev.type == ALLEGRO_EVENT_KEY_DOWN) {
               if (ev.keyboard.keycode == ALLEGRO_KEY_ESCAPE) {return;}
               switch (ev.keyboard.keycode) {
                  case ALLEGRO_KEY_UP :
                     
                     break;
                  case ALLEGRO_KEY_DOWN :
                     
                     break;
                  case ALLEGRO_KEY_LEFT :
                     
                     break;
                  case ALLEGRO_KEY_RIGHT :
                     
                     break;
                  default :
                     
                     break;
               }
            }
            
         } while (!al_is_event_queue_empty(queue));
      }
   }
   
///   POINT ScreenCoordsToIsoCoords(POINT p);
///   POINT IsoCoordsToScreenCoords(POINT p);
   POINT ScreenCoordsToIsoCoords(POINT scr) {
      scr -= isoWorldPos;
      scr.y = -scr.y;
      POINT iso;
      iso.x = scr.x/ISOTILEWIDTH + scr.y/ISOTILEHEIGHT;
      iso.y = scr.y/ISOTILEHEIGHT - scr.x/ISOTILEWIDTH;
      return iso;
   }
   POINT IsoCoordsToScreenCoords(POINT iso) {
      POINT scr;
      scr.x = (ISOTILEWIDTH/2.0f)*(iso.x - iso.y);
      scr.y = (ISOTILEHEIGHT/2.0f)*(iso.x + iso.y);
      scr.y = -scr.y;
      scr += isoWorldPos;
      return scr;
   }
};

int main(int argc , char** argv) {
   IsometricWorld* world = 0;
   
   try {
      world = new IsometricWorld();
      world->Run();
   }
   catch (GameException e) {
      printf("Error - %s\n" , e.what());
   }
   
   return 0;
}

